-- This files containts basic components for WISHBOHE
-- author Ernest Jamro jamro@agh.edu.pl

---------------------------------------------------------------------------------------
-- fifo - first in first out 
-- maximum number of data stored inside the internal buffer is 2^adr_width-1
-- because only small distributed RAMs are employed the adr_width should be 4

-- the major task of this fifo is to accept new data continuesly 

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////

-- for adr_width= 1 use direct input - output connection
   --  adr_width= 2 - use only a single registers (do not use memory) 

	
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity fifo is
  generic( adr_width: integer:= 4; -- internal memory address width - 4 is preferable for distributed rams, 0- direct connection
	  data_width: integer:= 8); -- data bus width
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
	rst: in std_logic; -- synchronous reset for the whole device
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    half_full: out std_logic; -- fifo is half full
	-- data input signals (slave)
	datI_I: in std_logic_vector(0 to data_width-1);
	stbI_I: in std_logic; -- strobe similar as OPB_select (master ready to transfer data)
	ackI_O: out std_logic; -- acknowledge
	-- data output signals (master)
	datO_O: out std_logic_vector(0 to data_width-1);
	stbO_O: out std_logic;
	ackO_I: in std_logic);
end fifo;

architecture wb_fifo_wra of fifo is
-- find maximum value of two integers
function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
end MAX;

 constant awidth: integer:= MAX(1, adr_width); -- address width for signal declaration
  signal empty: std_logic; 
  signal count_in, count_out: std_logic_vector(awidth-1 downto 0); -- counter where place data in into the memory or from where read data out
  signal count_dif: std_logic_vector(awidth-1 downto 0); -- counter how many data are inside the buffer = count_in-count_out
  signal ackI: std_logic; -- buffer is full and cannot accept new data, or is ampty and cannot send new data
  
  type ram_type is array ((2**awidth)-1 downto 0) of std_logic_vector (0 to data_width-1); 
  signal RAM : ram_type; 
  signal we: std_logic; -- memory write enable
begin
 gadr1: if adr_width>= 1 generate
 -- input counter
 process(clk, arst) begin
   if arst='1' then count_in<=(others=>'0');
   elsif clk'event and clk='1' then
	 if rst='1' then count_in<= (others=>'0');
	 elsif ackI='1' then
		 count_in<= count_in + 1;
	 end if;
   end if;
 end process;
  ackI_O<= ackI;
  ackI<= not full and stbI_I;

 --  counter to address data_out of the dual port memory part
 process(clk, arst) begin
   if arst='1' then count_out<=(others=>'0');
   elsif clk'event and clk='1' then
	 if rst='1' then count_out<= (others=>'0');
	 elsif ackO_I='1' then
		 count_out<= count_out + 1;
	 end if;
   end if;
 end process;
 stbO_O<= not empty;

 -- fifo counter - how many data are in the fifo (count_dif)
 -- count_dif<= count_in - count_out; 
 c_dif: process(clk, arst) begin
   if arst='1' then count_dif<= (others=>'0');
   elsif clk'event and clk='1' then 
    if rst='1' then count_dif<= (others=>'0');
    elsif ackI='1' and ackO_I='0' then
      count_dif<= count_dif + 1;
    elsif ackI='0' and ackO_I='1' then
      count_dif<= count_dif - 1;
    end if;
   end if;
 end process;
 -- empty logic
 process(clk, arst) begin
   if arst='1' then empty<= '1';
   elsif clk'event and clk='1' then
     if rst='1' then empty<= '1';
     elsif ackI='1' then -- new data are transfered in
		 empty<= '0';
	 elsif count_dif=1 and ackO_I='1' then -- only one data in the RAM and now this data will be transffered out
		 empty<='1';
	 end if;
   end if;
 end process;
 almost_full<= '1' when count_dif = (2**adr_width) - 1 or full='1' else '0';
 almost_empty<= '1' when count_dif = 1 or empty='1' else '0';
 half_full<= count_dif(adr_width-1);
 
 -- full logic - full should be '1' whenever the count_dif = 2**adr_width-1
 process(clk, arst) begin
   if arst='1' then full<= '0';
   elsif clk'event and clk='1' then
	 if rst='1' then full<= '0';
     elsif (full='1' and ackO_I='0') or -- already full and cannot send data out (do not accept new data when full)
		 ( count_dif=(2**adr_width)-1 and ackO_I='0' and ackI='1' ) then -- almost full and new data will be accepted in and no data send out
	   full<= '1';
	 else full<= '0';
	 end if;
   end if; 
  end process;  

 -- internal memory
 we<= ackI; -- write enable
 process (clk) 
 begin 
 	if (clk'event and clk = '1') then  
 		if (we = '1') then 
 			RAM(conv_integer(count_in)) <= datI_I; 
 		end if; 
 	end if; 
 end process;
 
 --if adr_width<= 5 generate -- use distributed RAMs
   datO_O <= RAM(conv_integer(count_out));
 --end generate;
 
 end generate; -- adr_width>= 1
 -----------------------------------------------------------------------------------------------------------
 
 gadrm1: if adr_width<=0 generate -- use direct connection
     datO_O<= datI_I;
     stbO_O<= stbI_I;
     ackI_O<= ackO_I;
     almost_full<= '1';
     almost_empty<= '1';
     half_full<= '0';
     full<= '0'; -- fifo is never full 
 end generate;
 
 assert adr_width<=5
   report "This architecture does not support Block RAM, therefore distributed RAMs are employed for such a large FIFO"
	  severity warning; 

 
end wb_fifo_wra;
