-- plik interface urzadzenia slave - kazde urzadzenie slave powinno korzystac z tego modulu

-- Author: Ernest Jamro (AGH University of Science and Technology - Cracow Poland)
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement and author name is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////
--//////////////////////////////////////////////////////////////////////


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity OPB_slave is
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_REAL_DWIDTH: integer:= 0; -- real OPB bus data width used during data transfer (=0 - same as dwidth)
    DWIDTH : integer:= 8; -- user width of the data - the rest LSBs are always 0s
    AWIDTH : integer:= 16; -- user address width
    C_BASEADDR   : std_logic_vector          := X"0000_0000";
    C_HIGHADDR   : std_logic_vector          := X"0000_FFFF";
	  C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
	   -- and the lower occupied area however the lower clock frequency is required. 
     -- Therefore to increase the clock frequency this parameter should be increased
	  -- user interface data and address width
    fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
    fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
    use_toutSup: integer:= 0; -- >0 generate timeout Suppress signal whenever this slave is selected
    use_be: integer:= 1; -- use Byte Enable logic;  <=0 - Byte enable are always '1' (only full data transfres are used)
    user_full_transfer: integer:= 1; -- user side full data bus transfer, the parameter is used for  dwidth!= K*real_dwidth or real_dwidth!= K*dwidth
    USE_SeqAddr : integer:= 1 -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
    );
  port ( clk : in std_logic:= '0'; -- clock signal
    arst: in std_logic:= '0'; -- asynchronous reset (needed only for simulation)
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_rst     : in std_logic:= '0';
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- user signals
	  sel: out std_logic; -- select - active when OPB_select and proper address
	  ack: in std_logic; -- acknowlenge - slave is ready for data transfer
    retry: in std_logic:= '0'; -- generate retry signal
	  rnw: out std_logic; -- read not write - same assert OPB_read not write
	  adr: out std_logic_vector(0 to awidth-1); -- address bus
	  drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	  dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	  be: out std_logic_vector(0 to (dwidth+7)/8 -1); -- byte enable 
	  seqAddr: out std_logic -- same as OPB_seqAddr
    );
end OPB_slave;


architecture opb_slave_arch of OPB_slave is
 component opb_master
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width, the LSBs of OPB_DBus='0'
    real_dwidth: integer:= 0; -- opb bus dwidth (=0 then the same as dwidth)
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal user address bus width
    transfer_size: integer:= 4; -- max transfer size when bus is locked (valid only when seqAddr>0)
    use_be: integer:= 1; -- >0 use Byte Enable Logic, <=0 assume full dwidth transfer all the time
    use_seqAddr: integer:= 1; -- =1 use sequential addressing, =0 -ignore seqAddr signal
    use_addr_counter: integer:= 0; -- <=0 - address is loaded with select signal, >0 use internal address counter, in this mode address (adr) is loaded when load_addr='1' (sel signal is ignored for address logic) and is internally increamented after every data transfer, for reading operation two options are available =1 start reading after load_addr='1', =2 start reading just after reset signal from address=0
    single_master: integer:= 0; -- =1 only a single master is atached to the OPB bus - seting this parameter allows for additional hardware savings, =0 more than one master is connected to the OPB bus
    user_full_transfer: integer:= 1; -- user side full data bus transfer, the parameter is used for  dwidth!= K*real_dwidth or real_dwidth!= K*dwidth
    fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
    fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
    wr_pipeline_latency: integer:= 0); -- user (external) write logic pipeline latency, (=0 - combinational logic) 1- only a single pipeline register - dwr are delayed by 1 clock in comparison to the address, ack- signal is ack for the address - data (di) are accepted after 1 clk latter (stb and ack signal is disregarded)
 port(clk, arst: std_logic; --clock and global asynchronous reset
		-- OPB Master interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO (opb_dwidth+7)/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    M_request : out STD_LOGIC;
      OPB_rst: in std_logic:= '0';
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
      -- internal simplified bus
      load_addr: in std_logic:= '0'; -- load internal address counter - signal is ignored when use_addr_counter<=0
      fifo_full, fifo_empty: out std_logic; -- fifo is full and empty
      fifo_almost_full, fifo_almost_empty: out std_logic; -- only one data can be transferred and fifo will be full or empty
      retry: out std_logic; -- FIFO state indicates that it is better to dwaw retry signal (this signal might be used by an opb_slave module connected to this opb_master module)
      di: in std_logic_vector(0 to dwidth-1):= (others=>'0'); -- data input
      do: out std_logic_vector(0 to dwidth-1); -- data output
      be: in std_logic_vector(0 to (dwidth+7)/8 -1):= (others=>'1'); -- byte enable
      a: in std_logic_vector(0 to awidth-1); -- address bus
      msb_adr: in std_logic_vector(0 to opb_awidth-1):=(others=>'0'); -- the MSB of the OPB address are specified by the msb_adr
      rnw: in std_logic; -- read not write
	    sel: in std_logic; -- OPB select signal
      seqAddr: in std_logic; -- seqAddressing - a (address) is incremented - block transfer - similar logic assert OPB_seqAddr
	    ack: out std_logic);
end component;

function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
  if arg=0 then
    return 0;
  end if;
  for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "Address space is not a power of 2"
	severity error;
	return -1;
end log2;

function log2(arg: in std_logic_vector) return integer is
  begin
    for i in arg'LENGTH-1 downto 0 loop
      if arg(i)='1' then
        if i/=0 and arg(i-1 downto 0) /= 0 then
          assert false
          report "Address space is not a power of 2"
	        severity error;
          return -1; --error
        else 
          return i;
        end if;
      end if;
    end loop;
    return arg'LENGTH; -- for all zeros
end log2;

function FindRealDwidth(ignore: integer) return integer is begin
  if c_real_dwidth<=0 then
    return dwidth;
  else
    return c_real_dwidth;
  end if;
end FindRealDwidth;
  constant real_dwidth: integer:= FindRealDwidth(0);

  --constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant this_awidth1: integer:= log2(c_highaddr - c_baseaddr + 1);
  constant this_awidth: integer:= log2(c_highaddr - c_baseaddr + 1) - log2(8/real_dwidth) + log2(8/dwidth); -- address width (slave)
  constant mawidth: integer:= awidth + log2(8/real_dwidth) - log2(8/dwidth); -- master address width
  
  signal my_adr: std_logic; -- this device is selected (proper address)
  -- signals between slave interface and opb_master interface
  signal ssel, srnw: std_logic;
  signal sbe: std_logic_vector(0 to (real_dwidth-1)/8);
  signal sl_DBus1: std_logic_vector(0 to real_dwidth-1); -- slave data to be read
  -- signal between opb_master and user interface  
  signal fifo_retry: std_logic;
begin  
  adr_all: if awidth = c_opb_awidth generate -- the device is always selected
    my_adr<= '1';
    ssel<= OPB_select;
    srnw<= OPB_RNW;
    sbe<= OPB_BE(0 to (real_dwidth-1)/8);
  end generate;
  
  an: if awidth < c_opb_awidth generate -- decode address space
   speed0: if c_speedrate=0 generate 
      my_adr<= '1' when OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
         else '0';
    
     ssel<= OPB_select and my_adr;
     srnw<= OPB_rnw;
     sbe<= OPB_BE(0 to (real_dwidth-1)/8);
   end generate;

   speed1: if c_speedrate=1 generate
      process(arst, clk) begin
	      if arst='1' then my_adr<='0'; 
        elsif clk'event and clk='1' then
		      if OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
	          and OPB_select='1' and (ack='0' or (OPB_seqAddr='1' and use_seqAddr>0)) then
		        my_adr<= '1';
		      else my_adr<='0';
	        end if;  
	      end if;
	    end process; 
    ssel<= OPB_select and my_adr;
    srnw<= OPB_rnw;
    sbe<= OPB_BE(0 to (dwidth-1)/8);
  end generate; -- decode address space 
  
  speed2: if c_speedrate=2 generate
      process(arst, clk) begin
	      if arst='1' then ssel<= '0'; srnw<= '0'; sbe<= (others=>'0');
	      elsif clk'event and clk='1' then
	        srnw<= OPB_rnw;
          sbe<= OPB_BE(0 to (dwidth-1)/8);
          if OPB_ABus(0 to C_OPB_AWIDTH - awidth - 1) = C_BASEADDR(0 to C_OPB_AWIDTH - awidth - 1)
	          and (opb_select='1' and ack='0') then
		         ssel<= '1';
		      else 
             ssel<='0';
	        end if;  
	      end if;
	    end process; 
  end generate;  -- speed 2

end generate; -- decode address space 
  

  Sl_toutSup<= '0' when use_toutSup<= 0 else ssel;
  gd: if c_opb_dwidth>real_dwidth generate -- different data width for OPB_bus and user bus
	   Sl_DBus(real_dwidth to c_opb_dwidth-1)<= (others=>'0');
  end generate;
  Sl_errAck<= '0';

  -- do not use opb_maser interface  
  m0: if real_dwidth=dwidth and fifo_wr_awidth<=0 and fifo_rd_awidth<=0 and use_seqAddr<2 generate
    sel<= ssel;
    rnw<= srnw;
    be<= sbe  when use_be>0 else (others=>'1');
    Sl_xferAck<= ack;
    Sl_retry<= ssel and retry;
    Sl_DBus(0 to real_dwidth-1)<= drd when ssel='1' and srnw='1' and fifo_rd_awidth>=0 else (others=>'0');
    dwr<= OPB_DBus(0 to dwidth-1);
    seqAddr<= OPB_seqAddr;
    adr<= OPB_ABus(c_opb_awidth-awidth to c_opb_awidth-1);
  end generate;
  
  -- use OPB master interface
  m1: if real_dwidth/=dwidth or fifo_wr_awidth>0 or fifo_rd_awidth>0 or use_seqAddr>1 generate
     master: opb_master
     generic map(opb_dwidth=> dwidth, dwidth=> real_dwidth, real_dwidth=> dwidth,
       opb_awidth=> awidth, awidth=> mawidth,
       transfer_size=> 1, use_be=> use_be, user_full_transfer=> 1 - user_full_transfer,
       use_seqAddr=> use_seqAddr, use_addr_counter=>0,
       single_master=> 1, fifo_wr_awidth=> fifo_wr_awidth, fifo_rd_awidth=> fifo_rd_awidth,
       wr_pipeline_latency=> 0)
     port map(clk=> clk, arst=>arst,
		-- OPB Master interface
	    M_ABus=> adr, M_BE=> be, M_RNW=> rnw, M_select=> sel,
	    M_seqAddr=> seqAddr, M_busLock=> open, M_DBus=> dwr, M_request=> open,
      OPB_rst=> OPB_rst, OPB_MGrant=> '1', OPB_pendReq=> '0', OPB_DBus=> drd, 
	    OPB_errAck=> '0', OPB_retry=> retry, OPB_timeout=> '0', OPB_xferAck=> ack,
      -- internal simplified bus
      load_addr=> '0', fifo_full=> open, fifo_empty=> open,
      fifo_almost_full=> open, fifo_almost_empty=> open, retry=> fifo_retry,
      di=> OPB_DBus(0 to real_dwidth-1), do=> Sl_DBus1,
      be=> sbe, a=> OPB_ABus(c_opb_awidth-mawidth to c_opb_awidth-1),
      rnw=> srnw, sel=> ssel, seqAddr=> OPB_seqAddr, ack=> Sl_xferAck);
   
     Sl_retry<= ssel and (fifo_retry or retry);
     Sl_DBus(0 to real_dwidth-1)<= Sl_DBus1 when ssel='1' and srnw='1' and fifo_rd_awidth>=0 else (others=>'0');
  end generate;
  

    
    
  assert this_awidth = awidth
   report "opb_slave module error. Different address width for awidth and address space"
   severity error;
   
  assert not (c_speedrate<0 or c_speedrate>=3)
    report "opb_slave. The c_speedrate value must be in the range 0 to 2."
  	severity error;
	
  process(clk) begin
    if clk'event and clk='1' then
      
      assert not (ssel='1' and real_dwidth<=8 and 
       (OPB_BE(1)='1' or OPB_BE(2)='1' or OPB_BE(3)='1'))
       report "opb_slave. Wrong real_dwidth settings, slave data width is <= 8 but master real data width is greater than 8"
       severity warning;
    
      assert not ( ssel='1' and real_dwidth>8 and real_dwidth<=16 and 
         (OPB_BE(2)='1' or OPB_BE(3)='1'))
        report "opb_slave. Wrong real_dwidth settings, slave data width is 9..16 but master real data width is greater than 16"
        severity warning;
      
      assert not (ssel='1' and real_dwidth>8 and 
         (OPB_BE(0)='1' and OPB_ABUS(C_OPB_AWIDTH-1)='1'))
        report "opb_slave. Wrong real_dwidth settings, slave data width is >8 but master real data width is <= 8"
        severity warning;
        
      assert not (ssel='1' and real_dwidth>16 and 
        ( (OPB_BE(0)='1' and OPB_ABUS(C_OPB_AWIDTH-2 to C_OPB_AWIDTH-1)/="00") or 
          ( OPB_BE(0 to 1)= "01" and OPB_ABUS(C_OPB_AWIDTH-2 to C_OPB_AWIDTH-1)/="01") or
          (OPB_BE(0 to 2)= "001" and OPB_ABUS(C_OPB_AWIDTH-2 to C_OPB_AWIDTH-1)/="10") or
          (OPB_BE(0 to 3)= "0001" and OPB_ABUS(C_OPB_AWIDTH-2 to C_OPB_AWIDTH-1)/="11")  ))
        report "opb_slave. Wrong real_dwidth settings, slave real data width is >16 but master real data width is <= 16"
        severity warning;
        
    end if;
  end process;
    
    
   
end opb_slave_arch;